/*

BYOND Hub Module

You can use this code library to report scores (ie wins/losses) to a central
score file maintained by the hub.  A web page of player ranks is produced
from this and made accessible through your game's hub page.

You can also generate shared savefiles which are stored at the hub, and
which are therefore called "hubfiles".  The users marked as owners of
the file may check it out, make a move, and then check it back in.

*/

#include "hubmsg.dm"
#include "hubinfo.dm"

var/ByondHub/byondhub = new() //global byondhub object

#define GENERAL_HUB_INFO 1
#define WORLDS_HUB_INFO  2
#define USERS_HUB_INFO   4

ByondHub

	proc
		ReportScores(scores[])
		GetRanks(players[])
		CheckinHubfile(File,users[],title)
		DeleteHubfile()
		SendEmail(recipient,subject,message) //this must be a persistent world accessible via world.hub because the hub rings us back to double-check
		GetHubInfo(hub=world.hub,content) //returns a /HubInfo object; content = null or any of _HUB_INFO constants ORed together

		//hooks (ie things you can override to change the user interface)
		ConfirmScores(scores[])
		HandleFailedConnection()
		HandleFailedQuery()
		HandleFailedCheckout()
		HandleFailedCheckin()
		LoadHubfile(File,users[],title,resolved_emails[])

		//low level procs
		HubWebPage(hub=world.hub)
		SendQuery(qparams[])
		PreAuthorizeMsg(qparams[])
		AuthorizeMsg(qparams[])

	var/const
		USER_ACTIVE   = 1
		USER_INACTIVE = 2

	var
		hub_services_address = "byond://Dantom.HubServices"
		hub_web_address = "http://www.byond.com/hub/hub.cgi"
		hub_file_id
		auth_msgs[0]

ByondHub/New()
	tag = BYOND_HUB_TAG  //must match tag used by the hub to direct Topic() messages to this object
	return ..()

ByondHub/HandleFailedConnection()
	world.log << "Failed to connect to [hub_services_address]."

ByondHub/HandleFailedQuery(msg)
	world.log << "Hub query failed: [msg]"

ByondHub/SendQuery(qparams[])
	var/result = world.Export("[hub_services_address]?[list2params(qparams)]")
	if(result)
		qparams = params2list(result)
		if(!qparams)
			qparams = params2list(RESULT_PRM + "=" + result)
	if(!qparams)
		HandleFailedConnection()
		return
	if(qparams[RESULT_PRM] == ERROR_RSLT)
		HandleFailedQuery(qparams[MSG_PRM])
		return
	qparams -= RESULT_PRM
	return qparams

ByondHub/HubWebPage(hub=world.hub)
	if(!hub) return
	return "[hub_web_address]?qd=hub;hub=[hub]"

ByondHub/GetHubInfo(hub=world.hub,content)
	if(!hub) return
	var/url = HubWebPage(hub) + ";format=text"
	if(content)
		url += "content=";
		if(content & GENERAL_HUB_INFO) url += "|general|"
		if(content & WORLDS_HUB_INFO) url += "|worlds|"
		if(content & USERS_HUB_INFO) url += "|users|";

	var/http_result = world.Export(url)
	if(!http_result) return

	var/txt = http_result["CONTENT"]
	if(!txt) return

	var/savefile/F = new
	F.ImportText("/",file2text(txt))

	var/HubInfo/hub_info = new
	if("general" in F.dir)
		F.cd = "general"
		hub_info.Read(F)
		F.cd = ".."

	hub_info.worlds = list()

	if("worlds" in F.dir)
		F.cd = "worlds"
		for(var/w in F.dir)
			var/HubWorldInfo/world_inf = new
			F.cd = w
			world_inf.Read(F)
			hub_info.worlds += world_inf
			F.cd = ".."
		F.cd = ".."

	return hub_info

ByondHub/ConfirmScores(scores[])
	var/p

	world << "<span class=hub>Adding scores to the <a href=[HubWebPage()]>rank list</a>:</span>"
	for(p in scores)
		world << "<span class=hub>[p] [scores[p]]</span>"

	//by default, we automatically accept these scores
	return 1

ByondHub/GetRanks(players[])
	var/qparams[0]
	var/p
	for(p in players)
		if(ismob(p) || istype(p,/client)) qparams += p:key
		else if(istext(p)) qparams += p
		else world.log << "ByondHub: invalid player entry: [p]"

	if(!length(qparams)) return
	if(!world.hub || world.hub < 0)
		world.log << "You must have assigned world.hub in order to get ranks."
		return

	qparams[QUERY_PRM] = GET_RANKS_QRY
	qparams[HUBID_PRM] = world.hub

	return SendQuery(qparams)

ByondHub/ReportScores(scores[])
	var/qparams[0]
	var/p
	for(p in scores)
		var/s = scores[p]
		if(ismob(p) || istype(p,/client)) qparams[p:key] = s
		else if(istext(p)) qparams[p] = s
		else world.log << "Invalid score entry: [p]"

	if(!ConfirmScores(qparams)) return
	if(!length(qparams)) return
	if(!world.hub)
		world.log << "You must have assigned world.hub in order to report scores."
		return

	qparams[QUERY_PRM] = ADD_SCORES_QRY
	qparams[HUBID_PRM] = world.hub
	if (world.byond_version >= 345 && world.hub_password)
		qparams[HUBPW_PRM] = world.hub_password

	return SendQuery(qparams)

ByondHub/HandleFailedCheckout()
	world.log << "Failed to checkout hubfile."
ByondHub/HandleFailedCheckin()
	world.log << "Failed to checkin hubfile."


world/Topic(href,address)
	var/href_list[] = params2list(href)
	var/ByondHub/hub = locate(href_list["src"])
	if(istype(hub))
		href_list[REMOTE_ADDR_PRM] = address
		return hub.Topic(href,href_list)
	else return ..()

ByondHub/proc/ExtractEmail(header)
	//expected format of header: "key <email>" or "key" or "<email>"
	var/email_start = findtext(header,"<")
	if(!email_start) return null

	email_start++
	var/email_end = findtext(header,">",email_start)
	return copytext(header,email_start,email_end)

ByondHub/proc/ExtractKey(header)
	//expected format of header: "key <email>" or "key" or "<email>"
	var/key_end = findtext(header,"<")
	if(!key_end) return header

	//trim trailing whitespace
	while(key_end > 1 && findtext(header," ",key_end-1,key_end))
		key_end--

	return copytext(header,1,key_end)

ByondHub/Topic(href,href_list[])

	switch(href_list[QUERY_PRM])
		if(LOAD_HUBFILE_QRY)
			var/users[0]
			var/resolved_emails[0]
			var p, email, key

			//one or more players may be in the "wait" and "turn" lists
			for(p in list() + href_list["wait"])
				if(p)
					email = ExtractEmail(p)
					key = ExtractKey(p)
					users[key || email] = USER_INACTIVE
					if(key && email) resolved_emails[email] = key
			for(p in list() + href_list["turn"])
				if(p)
					email = ExtractEmail(p)
					key = ExtractKey(p)
					users[key || email] = USER_ACTIVE
					if(key && email) resolved_emails[email] = key

			hub_file_id = href_list["file"]
			var/title = href_list["title"]

			var/import_qry[0]
			import_qry["file"] = hub_file_id
			import_qry[QUERY_PRM] = LOAD_HUBFILE_QRY
			var/File = usr.client.Import("\ref[src]" + list2params(import_qry))

			if(!File) HandleFailedCheckout()
			else LoadHubfile(File,users,title,resolved_emails)

		if(AUTH_EMAIL_QRY)
			return AuthorizeMsg(href_list)

ByondHub/DeleteHubfile()
	if(hub_file_id)
		CheckinHubfile(null)

ByondHub/CheckinHubfile(File,users[],title)
	var/qparams[0]
	var/wait[0]
	var/turn[0]
	var/p
	for(p in users)
		var/status = users[p]
		if(ismob(p) || istype(p,/client)) p = p:key
		else if(!istext(p))
			world.log << "Invalid user entry: [p]"
			continue

		if(findtext(p,"@"))
			p = "<[p]>" //hub expects email addresses to be enclosed in <>

		if(status == USER_ACTIVE)
			turn += p
		else if(status == USER_INACTIVE)
			wait += p
		else
			world.log << "ByondHub Error: Invalid status for [p]."

	if(hub_file_id) qparams["file"] = hub_file_id
	if(length(turn)) qparams["turn"] = turn
	if(length(wait)) qparams["wait"] = wait
	if(title) qparams["title"] = title

	if(!usr || !usr.client)
		world.log << "ByondHub Error: SaveHubfile must be called by active usr"
	else
		while(1)
			var/url = "byond://[world.hub]?[list2params(qparams)]##action=hubfile"
			. = usr.client.Export(url,File)
			if(istext(.)) //param text response from BYOND Central (otherwise it is just a simple true/false numeric result)
				var/params[] = params2list(.)
				. = text2num(params["ok"])
				if(!.)
					var/checked_out = params["out"]
					var/checked_in = params["in"]
					var/yesno
					if(checked_out) yesno = alert("[checked_out] has this file checked out.  Are you sure you want to save it?","Override [checked_out]?","Yes","No")
					else if(checked_in) yesno = alert("[checked_in] has made changes to the hubfile since you downloaded it.  Overwrite those changes with yours?","Override [checked_in]?","Yes","No")
					if(yesno == "Yes")
						qparams["force"] = 1
						continue
			break
		if(!.) HandleFailedCheckin()

ByondHub/PreAuthorizeMsg(qparams[])
	for(var/i=1,i<1000,i++)
		var/msgid = "[rand(1,10000)]"
		if(!(msgid in auth_msgs))
			qparams[MSGID_PRM] = msgid
			auth_msgs[msgid] = qparams
			return 1

	var/global/warned
	if(!warned)
		warned = 1
		world.log << "ByondHub Error: exhausted message authorization queue."

ByondHub/AuthorizeMsg(qparams[])
	var/msgid = qparams[MSGID_PRM]

	if(msgid in auth_msgs)
		auth_msgs -= msgid
		return 1
	else
		world.log << "ByondHub Error: request to authorize message failed (remote IP = '[qparams[REMOTE_ADDR_PRM]]')!"

ByondHub/SendEmail(recipient,subject,message)
	var/qparams[0]
	var/sender
	var/hubpath = world.hub

	if(!hubpath)
		world.log << "ByondHub Error: world.hub must be configured before sending mail."
		return

	var/pos = findtext(hubpath,".")
	if(pos) sender = copytext(hubpath,1,pos)
	else
		world.log << "ByondHub Error: invalid world.hub setting '[hubpath]'."
		return

	qparams[QUERY_PRM] = SEND_EMAIL_QRY
	qparams[HUBID_PRM] = hubpath
	qparams[SENDER_PRM] = sender
	qparams[RECIPIENT_PRM] = recipient
	qparams[SUBJECT_PRM] = subject
	qparams[MSG_PRM] = message

	if(PreAuthorizeMsg(qparams))
		return SendQuery(qparams)
