/*

Hub Query Module

This module handles all communication with the hub.

*/

#include "hubmsg.dm"

HubQuery
	var
		hub_address = "byond://hub" //currently registered at "byond://dantom.com:6003"

		ping_period
		open_links[0]

	proc
		ShowUrls(path)     //display url list from LsQuery()
		ShowMatches(path)  //display directory names and urls from LsQuery()
		Shutdown()

		//basic message procs (no direct output)
		LsQuery(path)             //get directory listing
		OpenLink(path,desc,topic) //open a temporary portal from hub  (hub path, url description, local connection topic)
		CloseLink(path)      //call with full path (returned by OpenLink)
		PingLink(path)       //keep temporary portal alive

		//low-level procs
		SendQuery(qparams[])

		//hooks
		HandleFailedConnection()
		HandleFailedQuery(msg)

PortalInfo
	var
		name
		desc
		url
		cd

HubQuery/HandleFailedConnection()
	world.log << "Failed to connect to hub."
HubQuery/HandleFailedQuery(msg)
	world.log << "Hub query failed: [msg]"

HubQuery/SendQuery(qparams[])
	qparams = world.Export("[hub_address]#[list2params(qparams)]")
	if(qparams) qparams = params2list(qparams)
	if(!qparams)
		HandleFailedConnection()
		return
	if(qparams[RESULT_PRM] == ERROR_RSLT)
		HandleFailedQuery(qparams[MSG_PRM])
		return
	return qparams

HubQuery/ShowUrls(path)
	var/Results[] = LsQuery(path)
	if(Results)
		var/title
		var/PortalInfo/P
		for(P in Results)
			if(!P.url) continue

			if(!title)
				title = P.cd || "hub"
				usr << "Open portals in [title]:"

			if(!P.desc) P.desc = P.url
			if(P.url) usr << "<a href='[P.url]'>[P.desc]</a>"

		if(!title) usr << "There are no open portals."

HubQuery/ShowMatches(path)
	var/Results[] = LsQuery(path)
	if(Results)
		var/title
		var/PortalInfo/P
		for(P in Results)
			if(!title)
				title = P.cd || "hub"
				usr << "Contents of [title]:"

			if(!P.desc) P.desc = P.url
			if(P.url) usr << "[P.name] <a href='[P.url]'>[P.desc]</a>"
			else usr << "[P.name]"

		if(!title) usr << "No portals found."

HubQuery/LsQuery(path)
	var/qparams[0]
	qparams[QUERY_PRM] = LS_QRY
	qparams[PATH_PRM] = path

	qparams = SendQuery(qparams)
	if(!qparams || qparams[RESULT_PRM] != SUCCESS_RSLT) return

	var/Results[0]
	var/cd
	var/P
	cd = qparams[PATH_PRM]
	for(P in qparams)
		//filter out parameters from paths
		if(P == DESC_PRM)
			var/PortalInfo/I = new()
			I.name = cd //TODO: strip all but final name
			I.cd = cd
			I.desc = qparams[P]
			Results.Add(I)
			continue
		if(findtext(P,PARAM_TAG) == 1) continue

		var/PortalInfo/I = new()
		var/d = qparams[P]
		I.name = P
		I.cd = cd
		if(d)
			var/p = findtext(d,"\n")
			if(!p) I.url = d
			else
				I.url = copytext(d,1,p)
				I.desc = copytext(d,p+1)

			//convert local hub addresses to full address
			if(findtext(I.url,"#") == 1) I.url = "[hub_address][I.url]"
			else if(findtext(I.url,"byond://#") == 1)
				I.url = "[hub_address][copytext(I.url,10)]"

		Results.Add(I)

	return Results

HubQuery/OpenLink(path,desc,topic)
	var/qparams[0]
	if(!world.port && !world.OpenPort())
		world.log << "This world must be hosted on a network port to link with the hub."
		return
	qparams[QUERY_PRM] = OPEN_QRY
	qparams[PATH_PRM] = path
	if(desc) qparams[DESC_PRM] = desc
	if(topic) qparams[TOPIC_PRM] = topic

	qparams = SendQuery(qparams)
	if(!qparams) return
	var/ping_timer = text2num(qparams[PING_TIMER_PRM])
	path = qparams[PATH_PRM]
	open_links.Add(path)
	StartPinger(ping_timer)
	return path

HubQuery/CloseLink(path)
	if(!open_links.Remove(path)) return
	var/qparams[0]
	qparams[QUERY_PRM] = CLOSE_QRY
	qparams[PATH_PRM] = path

	return SendQuery(qparams)

HubQuery/PingLink(path)
	var/qparams[0]
	qparams[QUERY_PRM] = PING_QRY
	qparams[PATH_PRM] = path

	return SendQuery(qparams)

HubQuery/Shutdown()
	var/path
	for(path in open_links)
		CloseLink(path)

//sleeping shutdown stuff must be done in world.Del() to work
world/Del()
	var/HubQuery/h
	for(h)
		h.Shutdown()
	return ..()

HubQuery/proc/StartPinger(period)
	if(!ping_period)
		if(!ping_period || ping_period < period) ping_period = period
		spawn
			while(open_links.len)
				sleep(ping_period)
				if(!open_links.len) break

				var/qparams[0]
				qparams[QUERY_PRM] = PING_BATCH_QRY
				qparams.Add(open_links)
				qparams = SendQuery(qparams)

				if(!qparams)
					world.log << "Dropping all hub connections."
					open_links.len = 0
				else if(qparams[RESULT_PRM] == SUCCESS_RSLT)
					var/path //returns list of failed pings
					for(path in qparams)
						if(!findtext(path,PARAM_TAG))
							world.log << "Hub dropped connection to [path]."
							open_links.Remove(path)

			ping_period = 0
