/*
	-----------------------------------------------------------------

	Sets (/Set)

	-----------------------------------------------------------------

	A set is a powerful data structure that conveniently stores
	a set of unique elements in no particular order.  Sets are
	used to store a collection of elements that may need to be
	sorted.  Sets in this library feature many convenient set
	operations, such as unions, intersections, complements.  They
	also impelement operations such as subset and element checks.

	-----------------------------------------------------------------

	Set constants

		var/const/SET_UNIVERSAL = 1
			Send this value through the third argument of /Set/New()
			to define the set as a universe.

	-----------------------------------------------------------------

	Set Operations

		/proc/set_Intersection(Set/A, Set/B)
			Returns the intersection of A and B

		/proc/set_Union(Set/A, Set/B)
			Returns the union of A and B

		/proc/set_Complement(Set/A, Set/B)
		/proc/set_Complement(Set/A)
			Returns the complement of A and B.  If B is not
			supplied, then it will return the absolute complement of
			A, assuming that A has a Universal Set defined.  The
			function will crash if no Universal Set or an invalid
			Universal Set was assigned.

		/proc/set_isEqual(Set/A, Set/B)
			Returns 1 if A is equal to B; otherwise 0.  Two sets are
			equal if they have equivalent contents.

		/proc/set_isSubSet(Set/A, Set/B)
			Returns 1 if A is a subset of B; otherwise 0.

		/proc/set_isElement(O, Set/B)
			Returns 1 if O is an element of set B; otherwise 0.

		/proc/setIn(A, list/L, operator)
			Internal.  Returns the reference of the object comparing
			to A if A is in the list L using 'operator' as the
			comparison operator.

		/proc/set_canOperate(Set/A, Set/B)
			Returns 1 if two sets have the same universal set and
			comparison function; otherwise 0.
			Two sets cannot operate on each other if they have
			inconsistant traits, therfore this function should be
			used to check before using any set operations.

	-----------------------------------------------------------------

	Set Comparison procs

		/proc/setEquals(A, B) (default value for /Set/var/comp)
			Returns 1 if A is equal to B (A is exactly the same as B)
			otherwise 0.

		/proc/setSameType(datum/A, datum/B)
			Returns 1 if A is either equal to B or if A has the
			same path type as B; otherwise 0.

	-----------------------------------------------------------------

	/Set
		New(list/contents, compare=/proc/setEquals, Set/univ=null)
			Constructs a new set given the optional initial contents,
			comparison function, and universal set.  The comparison
			function must have two arguments, returning true if A and
			B are equivalant in terms of the set.  'univ' is used to
			identify the universe for that particular set when using
			the absolute complement operation.


		/var/list/contents = null
			Internal list of elements in the set.  Do not touch.

		/var/comp = /proc/setEquals
			The comparison function.  Do not touch.


		/var/Set/UnivSet = null
			The universe of the set.  Do not touch.

		/proc/isMember(A)
			Returns 1 if A is a member of the set; otherwise 0

		/proc/getSize()
			Returns the size of the set.

		/proc/isNull()
			Returns 1 if the set is a null set (when there are no
			items in the set); otherwise 0

		/proc/Insert(A)
			Inserts A into the set.  Returns 0 if an equivalent
			element is already in the set; otherwise 1.

		/proc/Delete(A)
			Removes A from the set.  Returns 0 if an equivalent
			element is not founc in the set; otherwise 1.

		/proc/toString()
			Returns the contents of the set as a text string
			surrounded by braces.  Order is insignificant.

		/proc/isEqual(Set/S)
			Returns 1 if two sets are equal; otherwise 0.

		/proc/isSubSet(Set/B)
			Returns 1 if the set is a subset of B; otherwise 0.

		/proc/Intersect(Set/S)
			Internal.  Returns a list of the intersection of
			the set and S.

		/proc/Union(Set/S)
			Internal.  Returns a list of the union of the
			set and S.

		/proc/Complements(Set/B)
			Internal.  Returns the complement of the set and
			B.

		/proc/AbsComplement()
			Internal.  Returns a list of the absolute complement
			of the set.

	-----------------------------------------------------------------
*/

// Set constants
var/const
	SET_UNIVERSAL = 1

// Set operations
proc
	set_Intersection(Set/A, Set/B)
		var/list/L = A.Intersect(B)
		return new /Set (L, A.comp, A.UnivSet)

	set_Union(Set/A, Set/B)
		var/list/L = A.Union(B)
		return new /Set (L, A.comp, A.UnivSet)

	set_Complement(Set/A, Set/B)
		var/list/L = B ? A.Complements(B) : A.AbsComplement()
		return new /Set (L, A.comp, A.UnivSet)

	set_isEqual(Set/A, Set/B)
		return A.isEqual(B)

	set_isSubSet(Set/A, Set/B)
		return A.isSubSet(B)

	set_isElement(O, Set/A)
		return A.isMember(O)

	// Internal: is A in list L, using the comparison operator?
	setIn(A, list/L, operator)
		for(var/item in L)
			if(call(operator)(A, item))
				return item
		return null

	// returns 1 if two sets have the same universal set and comparison
	// function; otherwise 0
	// two sets cannot operate on each other if they have inconsistant traits
	set_canOperate(Set/A, Set/B)
		if(A.UnivSet == B)
			return (A.comp == B.comp)
		return (A.comp == B.comp) && (A.UnivSet == B.UnivSet)

// Set comparison procs
proc
	setEquals(A, B)
		return (A == B)

	setSameType(datum/A, datum/B)
		return setEquals(A, B) || (A.type == B.type)


Set
	var
		list/contents = null	// list of contents
		comp = /proc/setEquals	// comparison function
		Set/UnivSet = null		// universal set

	New(list/contents, compare=/proc/setEquals, Set/univ=null)
		src.comp = compare

		if(contents != null)
			if(istype(contents))
				src.contents = contents.Copy()
			else
				src.contents = list(contents)
		else
			src.contents = new /list

		if(univ)
			if(isSet(univ))
				src.UnivSet = univ
			if(univ == SET_UNIVERSAL)
				src.UnivSet = src

	proc/isMember(A)
		return setIn(A, src.contents, src.comp)

	proc/getSize()
		return length(src.contents)

	proc/isNull()
		return src.getSize() == 0

	proc/Copy()
		var/Set/S = new /Set (src.contents.Copy(), src.comp, src.UnivSet)
		return S

	proc/Insert(A)
		if(isMember(A))
			return 0
		src.contents += A
		return 1

	proc/Delete(A)
		if(!isMember())
			return 0
		src.contents -= A
		return 1

	proc/toString()
		var/text = "{ "
		for(var/item in src.contents)
			text += "\improper [item] "
		text += "}"
		return text

	// two sets equal when it has the same contents
	proc/isEqual(Set/S)
		if(!set_canOperate(src, S))
			upContainers_CRASH("Incompatible sets")

		if(src.getSize() != S.getSize())
			return 0

		var/list/temp = S.contents.Copy()
		for(var/i in src.contents)
			var/j = setIn(i, temp, src.comp)
			if(j != null)
				temp -= j
			else
				return 0

		return 1

	proc/isSubSet(Set/B)
		if(!set_canOperate(src, B))
			upContainers_CRASH("Incompatible sets")

		if(src.getSize() > B.getSize())
			return 0

		var/count = 0
		for(var/item in B.contents)
			if(item in src.contents)
				count++

		return (count == src.getSize())

	proc/Intersect(Set/S)
		if(!set_canOperate(src, S))
			upContainers_CRASH("Incompatible sets")

		var/list/L = list()
		for(var/item in src.contents)
			if(setIn(item, S.contents, src.comp))
				L += item
		return L

	proc/Union(Set/S)
		if(!set_canOperate(src, S))
			upContainers_CRASH("Incompatible sets")

		// using the law (a u b) = (a + b) - (a n b) would not be as fast
		// as manually adding the first set, and adding non-duplicate items from the
		// second set
		var/list/L = src.contents.Copy()
		for(var/K in S.contents)
			if(!setIn(K, L, src.comp))
				L += K
		return L

	proc/Complements(Set/B)
		if(!set_canOperate(src, B))
			upContainers_CRASH("Incompatible sets")

		var/list/L = B.contents.Copy()
		for(var/I in src.contents)
			var/K = setIn(I, L, src.comp)
			if(K != null)
				L -= K
		return L

	// absolute complement: complement of src in the Universal Set
	proc/AbsComplement()
		if(!src.UnivSet)
			upContainers_CRASH("Universal set does not exist")
		return src.Complements(src.UnivSet)



